/**
 * @file hl7800_client.cpp
 * @brief HL7800 modem TCP client API <br>
 * 
 * @date  2019/9/5 Preview-Alpha
 * @date  2019/9/18 Preview-Bata(bug fix)
 * @author kt-nakamura@kddi-tech.com
 */

#include "hl7800_client.h"
#include "hl7800_serial.h"
#include "hl7800_modem.h"

PROGMEM const char ATRSP_OK[]="OK\r";
PROGMEM const char ATRSP_ERR[]="ERROR\r";
PROGMEM const char ATRSP_KCNX_LF[]="\r";
PROGMEM const char ATDATA_EOF[]="--EOF--Pattern--";
PROGMEM const char ATRSP_CONNECT[]="CONNECT\r";
PROGMEM const char ATRSP_KTCP_IND1[]="+KTCP_IND: 1,1";
PROGMEM const char ATRSP_KTCP_IND0[]="+KTCP_IND: 1,0";

PROGMEM const char ATRSP_KTCP_DATA[]="+KTCP_DATA: 1,";
PROGMEM const char ATRSP_KTCP_NOTIF1[]="+KTCP_NOTIF: 1,4";

extern HL7800Serial serial_hl78;
// extern char* rcv_buf;

/**
 * TCPライブラリ初期化</BR>
 * @param なし
 * @return なし
 */
HL7800Client::HL7800Client() {
  recived_size = 0;
  connection_id = -1;
  tcp_connect = false;
  
}

/**
 * TCP接続開始(IP指定)<br>
 * @param ip IPアドレス
 * @param port TCPポート番号
 * @return  1:成功, 0:失敗
 */
//PROGMEM const char ATCMD_KTCPCFG_IP[]="AT+KTCPCFG=1,0,\"%d.%d.%d.%d\",%d\r";
//PROGMEM const char ATCMD_KTCPCNX[]="AT+KTCPCNX=1\r";
int HL7800Client::connect(IPAddress ip, uint16_t port) {
  // AT+KTCPCFG=1,0,"www.google.com",80
  // AT+KTCPCNX=1
  int ret =  0;
  char* rcv_buff;
  recived_size = 0;
  char long_cmd[50];

  sprintf_P(long_cmd,PSTR("AT+KTCPCFG=1,0,\"%d.%d.%d.%d\",%d\r"),ip[0],ip[1],ip[2],ip[3],port);
//  sprintf(long_cmd,ATCMD_KTCPCFG_IP,ip[0],ip[1],ip[2],ip[3],port);
//  DBGSERIAL.print(F("@@@@@ HL7800Client::connect(IP) cmd="));
//  DBGSERIAL.println(cmd);
  rcv_buff = serial_hl78.WriteCommandAndReadResponse(long_cmd, ATRSP_OK, ATRSP_ERR, MODEM_CONNECT_TIMEOUT);
  if (rcv_buff == NULL) {
    DBGSERIAL.println(F("@@@@@ TCP connect fail!"));
    return ret;
  }

  sprintf_P(cmd,PSTR("AT+KTCPCNX=%d\r"),1);
//  DBGSERIAL.print(F("@@@@@ HL7800Client::connect(IP) cmd="));
//  DBGSERIAL.println(cmd);
  rcv_buff = serial_hl78.WriteCommandAndReadResponse(cmd, ATRSP_OK, ATRSP_ERR, MODEM_CONNECT_TIMEOUT);
  if (rcv_buff == NULL) {
    return ret;
  }

  rcv_buff = serial_hl78.ReadResponse(ATRSP_KTCP_IND1, ATRSP_KTCP_IND0,  MODEM_LONG_TIMEOUT);  // TCP接続通知 / TCP切断通知
  if (rcv_buff == NULL) {
    DBGSERIAL.print(F("TCP connect timeout!"));
    // TCP切断/タイムアウトの場合受信バッファのゴミを捨てる
    rcv_buff = serial_hl78.ReadResponse("\r", ATRSP_ERR,  MODEM_SHORT_TIMEOUT);  
    return ret;
  }

//  DBGSERIAL.println(F("@@@@TCP Connect OK"));
  tcp_connect = true;
  ret = 1;
  return ret;
}

/**
 * TCP接続開始(host指定)<br>
 * @param host ホストFQDN文字列
 * @param port TCPポート番号
 * @return  1:成功, 0:失敗
 */
//PROGMEM const char ATCMD_KTCPCFG_HOST[]="AT+KTCPCFG=1,0,\"%s\",%d\r";
int HL7800Client::connect(const char *host, uint16_t port) {
  // AT+KTCPCFG=1,0,"www.google.com",80
  // AT+KTCPCNX=1
  int ret =  0; // failed
  char* rcv_buff;
  recived_size = 0;
  char long_cmd[strlen(host)+25];

  sprintf_P(long_cmd,PSTR("AT+KTCPCFG=1,0,\"%s\",%d\r"),host,port);
//  DBGSERIAL.print(F("@@@@@ HL7800Client::connect(host) cmd="));
//  DBGSERIAL.println(long_cmd);
  rcv_buff = serial_hl78.WriteCommandAndReadResponse(long_cmd, ATRSP_OK, ATRSP_ERR, MODEM_CONNECT_TIMEOUT);
  if (rcv_buff == NULL) {
    DBGSERIAL.println(F("@@@@@ TCP connect fail!"));
    return ret;
  }

  sprintf_P(cmd,PSTR("AT+KTCPCNX=%d\r"),1);
//  sprintf(cmd,ATCMD_KTCPCNX);
//  DBGSERIAL.print(F("@@@@@ HL7800Client::connect(host) cmd="));
//  DBGSERIAL.println(cmd);
  rcv_buff = serial_hl78.WriteCommandAndReadResponse(cmd, ATRSP_OK, ATRSP_ERR, MODEM_CONNECT_TIMEOUT);
  if (rcv_buff == NULL) {
    return ret;
  }

  rcv_buff = serial_hl78.ReadResponse(ATRSP_KTCP_IND1, ATRSP_KTCP_IND0,  MODEM_LONG_TIMEOUT);  // TCP接続通知 / TCP切断通知
  if (rcv_buff == NULL) {
    DBGSERIAL.print(F("TCP connect timeout!"));
    // TCP切断/タイムアウトの場合受信バッファのゴミを捨てる
    rcv_buff = serial_hl78.ReadResponse(ATRSP_KCNX_LF, ATRSP_ERR,  MODEM_SHORT_TIMEOUT);  
    return ret;
  }

//  DBGSERIAL.println(F("@@@@TCP Connect OK"));
  tcp_connect = true;
  ret = 1; // success connect
  return ret;
}

/**
 * TCPバイト列送信<br>
 * ※実際には呼び出されない<br>
 * @param data 送信データ
 * @return  送信済みデータバイト数
 */
size_t HL7800Client::write(uint8_t) {
  int ret = 0;
  DBGSERIAL.println(F("@@@@@ HL7800Client::write(char)"));
  return ret;
}

/**
 * TCPバイト列送信<br>
 * @param buf 送信データバッファ
 * @param size 送信サイズ
 * @return  送信済みデータバイト数
 */
size_t HL7800Client::write(const uint8_t *buf, size_t size) {
  // AT+KTCPSND=1,18
  int ret = 0;
  char* rcv_buff;
  char* send_buff = (char*)buf;
 // int send_size = strlen(send_buff);
  int send_size = size;
  recived_size = 0;

  // 未接続の場合はエラー
  if (!tcp_connect) {
    return -1;
  }

  // 入力データサイズチェック
//  if (send_size != size) {
//    DBGSERIAL.print(F("HL7800Client::write(buf) invalid data size!"));
//    return ret;
//  }

  DBGSERIAL.print(F("@@@@@ HL7800Client::write(buf) size="));
  DBGSERIAL.println(size);

//  char write_cmd[size + 21]; // 書き込みデータ＋”\r\n\r\n--EOF--Pattern--" の配列を確保する

  // 送信コマンド+KTCPSNDを発行し接続待ち
  sprintf_P(cmd,PSTR("AT+KTCPSND=1,%d\r\n"),size);
//  sprintf(cmd,ATCMD_KTCPSND,size);
//  DBGSERIAL.print(F("@@@@@ HL7800Client::write(buf) cmd="));
//  DBGSERIAL.println(cmd);
  rcv_buff = serial_hl78.WriteCommandAndReadResponse(cmd, ATRSP_CONNECT, ATRSP_ERR, MODEM_CONNECT_TIMEOUT);
  if (rcv_buff == NULL) {
    DBGSERIAL.println(F("@@@@@ TCP connect fail!"));
    return ret;
  }

  // TCP送信データ部分を送信
//  DBGSERIAL.println(F("@@@@@ HL7800Client::write(buf) buf="));
//  for (int i=0;i<size;i++) {
//    DBGSERIAL.write(buf[i]);
//  }
//  DBGSERIAL.println("_");
  serial_hl78.WriteData(buf, size);

  // 終端コード"--EOF--Pattern--"を発行し終了待ち
  rcv_buff = serial_hl78.WriteCommandAndReadResponse("--EOF--Pattern--", ATRSP_OK, ATRSP_ERR, MODEM_LONG_TIMEOUT);
  if (rcv_buff == NULL) {
    DBGSERIAL.println(F("@@@@@ TCP send fail!"));
    return ret;
  }

  ret = size;
  return ret;
}

/**
 * 受信データのチェック<br>
 * @param なし
 * @return  受信済みデータバイト数
 */
int HL7800Client::available() {
  int ret = -1;
  char* rcv_buff;

  // 未接続の場合はエラー
  if (!tcp_connect) {
    return ret;
  }

  // 前回受信したデータが残っている場合は残りデータ数を返す
  if ( recived_size > 0) {
    ret = recived_size;
    return ret;
  }

  // 受け取りデータが不在の場合は受信データ通知を待つ
//  DBGSERIAL.println(F("@@@@@ HL7800Client::available()"));
  // モデムからのデータ待ち
  // +KTCP_DATAを待つ
  rcv_buff = serial_hl78.ReadResponse(ATRSP_KTCP_DATA, ATRSP_KTCP_NOTIF1,  MODEM_LONG_TIMEOUT);  // 受信通知 / TCP切断通知
  if (rcv_buff == NULL) {
    DBGSERIAL.println(F("TCP Disconnect"));
    // TCP切断/タイムアウトの場合受信バッファのゴミを捨てる
    rcv_buff = serial_hl78.ReadResponse(ATRSP_KCNX_LF, ATRSP_ERR,  MODEM_SHORT_TIMEOUT);  
    return ret;
  }
  // レスポンスデータのみを受け取る
  rcv_buff = serial_hl78.ReadResponse(ATRSP_KCNX_LF, ATRSP_ERR,  MODEM_SHORT_TIMEOUT);  
  if (rcv_buff == NULL) {
    return ret;
  }
//  DBGSERIAL.print(F("@@@@@ data available size= "));
//  DBGSERIAL.println(rcv_buff);
  int size = atoi(rcv_buff);

  // AT+KTCPRCVを発行しモデムが受信したTCPデータを取得する
  sprintf_P(cmd,PSTR("AT+KTCPRCV=1,%d\r\n"),size);
//  DBGSERIAL.println(cmd);
  rcv_buff = serial_hl78.WriteCommandAndReadResponse(cmd, ATRSP_CONNECT, ATRSP_ERR, MODEM_CONNECT_TIMEOUT);
  if (rcv_buff == NULL) {
    DBGSERIAL.println(F("@@@@@ TCP read fail!(AT+KTCPRCV)"));
    return ret;
  }
  rcv_buff = serial_hl78.ReadResponse(ATDATA_EOF, ATRSP_ERR,  MODEM_CONNECT_TIMEOUT);  
  if (rcv_buff == NULL) {
    DBGSERIAL.println(F("@@@@@ TCP read fail!(EOF not found)"));
    return ret;
  }

  // EOF文字列は不要なので削除
  char* end = strstr_P(rcv_buff,ATDATA_EOF);
  *end = NULL;
  
  recived_size = strlen(rcv_buff);
  DBGSERIAL.print(F("@@@@@ recive data: size="));
  DBGSERIAL.println(recived_size);
//  DBGSERIAL.println(rcv_buff);

  // 実際に受信したデーター数を返す
  read_p = rcv_buff;
  ret = recived_size;
  return ret;
}

/**
 * TCP受信データ受け取り<br>
 * @param なし
 * @return  受け取り成功データバイト数
 */
int HL7800Client::read() {
  int ret = -1;
//  DBGSERIAL.println(F("@@@@@ HL7800Client::read(char)"));
  if (recived_size > 0) {
    ret = *read_p++;
    recived_size--;
  }
  return ret;
}

/**
 * TCP受信データ受け取り<br>
 * @param buf 受け取りデータバッファ
 * @param size 受信したいデータ数
 * @return 受け取り成功データバイト数
 */
int HL7800Client::read(uint8_t *buf, size_t size) {
  int ret = -1;
  DBGSERIAL.println(F("HL7800Client::read(buf)"));
  if (recived_size > 0) {
    if (size > recived_size) {
      ret = recived_size;
    } else {
      ret = size;
    }
    memcpy(buf,read_p,ret);
    recived_size = recived_size - ret;
    read_p = read_p + ret;
  }
  return ret;
}

/**
 * 受信バッファの先頭を取得する<br>
 * @param なし
 * @return  受信データ(NULL終端)
 */
int HL7800Client::peek() {
  int ret = -1;
  DBGSERIAL.println(F("@@@@@ HL7800Client::peek(buf)"));
  if (recived_size > 0) {
    ret = *read_p;
  }
  return ret;
}

/**
 * 送信完了待ち<br>
 * ※write()で終了待ちをしているため何もしない
 * @param なし
 * @return なし
 */
void HL7800Client::flush() {
  DBGSERIAL.println(F("@@@@@ HL7800Client::flush()"));
  recived_size = 0;
}

/**
 * TCP接続終了<br>
 * @param なし
 * @return なし
 */
void HL7800Client::stop() {
// AT+KTCPDEL=1
// AT+KTCPCLOSE=1,1
// +KCNX_IND: 1,5,30 --  <1>cnx cnf=1,<5>Idle time down counting started for disconnection, <idletime>30
  char* rcv_buff;
  recived_size = 0;

  tcp_connect = false;
  sprintf_P(cmd,PSTR("AT+KTCPCLOSE=%d,1\r"),1);
//  sprintf(cmd,ATCMD_KTCPCLOSE);
//  DBGSERIAL.print(F("@@@@@ HL7800Client::stop() cmd="));
//  DBGSERIAL.println(cmd);
  rcv_buff = serial_hl78.WriteCommandAndReadResponse(cmd, ATRSP_OK, ATRSP_ERR, MODEM_CONNECT_TIMEOUT);

//  sprintf(cmd,ATCMD_KTCPDEL);
  sprintf_P(cmd,PSTR("AT+KTCPDEL=%d\r"),1);
//  DBGSERIAL.print(F("@@@@@ HL7800Client::stop() cmd="));
//  DBGSERIAL.println(cmd);
  rcv_buff = serial_hl78.WriteCommandAndReadResponse(cmd, ATRSP_OK, ATRSP_ERR, MODEM_CONNECT_TIMEOUT);
  if (rcv_buff == NULL) {
    return;
  }

  DBGSERIAL.println(F("Success TCP disconnect"));
}

/**
 * TCP接続状態を取得<br>
 * @param なし
 * @return  1:接続中, 0:接続なし
 */
uint8_t HL7800Client::connected() {
//  DBGSERIAL.println(F("@@@@@ HL7800Client::connected()"));
  if (!tcp_connect) {
    return 0;
  }
  return 1;
}

/**
 * TCPクライアント互換<br>
 * @param なし
 * @return  1:接続中, 0:接続なし
 */
HL7800Client::operator bool() {
  return false;
}
