/**
 * @file hl7800_http.cpp
 * @brief HL7800 modem HTTP client API <br>
 * 
 * @date  2019/9/5 Preview-Alpha
 * @author kt-nakamura@kddi-tech.com
 */

#include "hl7800_http.h"
#include "hl7800_serial.h"
#include "hl7800_modem.h"

PROGMEM const char ATRSP_OK[]="OK\r";
PROGMEM const char ATRSP_ERR[]="ERROR\r";
PROGMEM const char ATDATA_EOF[]="--EOF--Pattern--";
PROGMEM const char ATRSP_CONNECT[]="CONNECT\r";

PROGMEM const char ATRSP_KHTTP_IND[]="+KHTTP_IND: 1,3,";

extern HL7800Serial serial_hl78;

/**
 * HTTPデータ応答を処理する<br>
 * "--EOF--Pattern--" までを受け取りbuffにコピー<br>
 * @param buff 受け取りデータバッファ
 * @return  受信データバイト数
 */
int parseData(char* buff) {
  int ret = -1;
  char* rcv_buff;

  rcv_buff = serial_hl78.ReadResponse(ATDATA_EOF, ATRSP_ERR,  MODEM_CONNECT_TIMEOUT);  
  if (rcv_buff == NULL) {
    DBGSERIAL.println(F("@@@@@ HTTP read fail!(EOF not found)"));
    return ret;
  }

  // EOF文字列は不要なので削除
  char* end = strstr_P(rcv_buff, ATDATA_EOF);
  *end = NULL;
  int recived_size = strlen(rcv_buff);
//  DBGSERIAL.print(F("@@@@@ recive data: size="));
//  DBGSERIAL.println(recived_size);
//  DBGSERIAL.println(rcv_buff);

  memcpy(buff,rcv_buff,recived_size);
  memset(buff+recived_size,0x0,1);
  return recived_size;
}

/**
 * HTTPレスポンス応答(+KHTTP_IND:)を処理する<br>
 * 文字列("200","404"を十進値に変換<br>
 * @param なし
 * @return  HTTPレスポンス応答値
 */
PROGMEM const char ATRSP_BS[]="\"";
 
int parseRespCode() {
  int ret = -1;
  char* rcv_buff;
  
  // +KHTTP_IND: 1,3,29,200,"OK"
  // <session_id>,<status>[,<data_len>,<st_code>,<st_reason>]
  // 行頭が来るまで待機
  rcv_buff = serial_hl78.ReadResponse(ATRSP_KHTTP_IND, ATRSP_ERR ,  MODEM_CONNECT_TIMEOUT);  
  if (rcv_buff == NULL) {
    DBGSERIAL.println(F("@@@@@ HTTP read fail!(HTTP resp not found)"));
    return ret;
  }
  
  // 行末が来るまで待機
  rcv_buff = serial_hl78.ReadResponse(ATRSP_BS, ATRSP_ERR,  MODEM_CONNECT_TIMEOUT);  
  if (rcv_buff == NULL) {
    DBGSERIAL.println(F("@@@@@ HTTP read fail!(HTTP resp not found)"));
    return ret;
  }

  // カンマ区切りを探す
  char* head = strstr(rcv_buff,",");
  if (head == NULL) {
    return ret;
  }
  char* end2 = strstr(++head, ",");
  if (end2 == NULL) {
    return ret;
  }
  *end2 = NULL;

  //文字列("200","404"を数値に変換し終了
  return atoi(head);
}

/**
 * HTTPライブラリ初期化<br>
 * @param なし
 * @return なし
 */
HL7800Http::HL7800Http() {
  http_connect = false;
}

/**
 * HTTPセッションの開始(IP指定)
 * @param ip IPアドレス
 * @param port TCPポート番号
 * @param http_version 0:HTTP, 2:HTTPS
 * @param login 認証ログイン名(10文字以内)
 * @param passwd　認証パスワード(10文字以内)
 * @return  1:成功, 0:失敗
 */
int HL7800Http::connect(IPAddress ip, uint16_t port, int http_version, char* login, char* passwd) {
  // AT+KHTTPCFG= <cnx cnf>, <http_server> [,<http_port> [,<http_version> [,<login> [,<password> [,<start> [,<af> [,<cipher_index>]]]]]]]
  // AT+KHTTPCNX=<session_id>

  char* rcv_buff;
  int ret =  0;
  if ((login != NULL) && (passwd != NULL)) {
    sprintf_P(cmd, PSTR("AT+KHTTPCFG=1,\"%d.%d.%d.%d\",%d,%d,\"%s\",\"%s\"\r"), ip[0],ip[1],ip[2],ip[3],port,http_version,login,passwd);
  } else {
    sprintf_P(cmd, PSTR("AT+KHTTPCFG=1,\"%d.%d.%d.%d\",%d,%d\r") ,ip[0],ip[1],ip[2],ip[3],port,http_version);
  }
  DBGSERIAL.print(F("@@@@@ HL7800Http::connect(IP) cmd="));
  DBGSERIAL.println(cmd);
  rcv_buff = serial_hl78.WriteCommandAndReadResponse(cmd, ATRSP_OK, ATRSP_ERR, MODEM_CONNECT_TIMEOUT);
  if (rcv_buff == NULL) {
    DBGSERIAL.println(F("@@@@@ HTTP connect fail!"));
    return ret;
  }
  http_connect = true;
  ret = 1;
  return ret;
}

/**
 * HTTPセッションの開始(URL指定)<br>
 * @param url URL文字列
 * @param port TCPポート番号
 * @param http_version 0:HTTP, 2:HTTPS
 * @param login 認証ログイン名(10文字以内)
 * @param passwd　認証パスワード(10文字以内)
 * @return  1:成功, 0:失敗
 */
int HL7800Http::connect(char* url, uint16_t port, int http_version, char* login, char* passwd) {
  // AT+KHTTPCFG= <cnx cnf>, <http_server> [,<http_port> [,<http_version> [,<login> [,<password> [,<start> [,<af> [,<cipher_index>]]]]]]]
  // AT+KHTTPCNX=<session_id>
  
  char* rcv_buff;
  int ret =  0;
  
  if ((login != NULL) && (passwd != NULL)) {
    sprintf_P(cmd, PSTR("AT+KHTTPCFG=1,\"%s\",%d,%d,\"%s\",\"%s\"\r") ,url,port,http_version,login,passwd);
  } else {
    sprintf_P(cmd, PSTR("AT+KHTTPCFG=1,\"%s\",%d,%d\r"),url,port,http_version);
  }
  DBGSERIAL.print(F("@@@@@ HL7800Http::connect(URL) cmd="));
  DBGSERIAL.println(cmd);
  rcv_buff = serial_hl78.WriteCommandAndReadResponse(cmd, ATRSP_OK, ATRSP_ERR, MODEM_CONNECT_TIMEOUT);
  if (rcv_buff == NULL) {
    DBGSERIAL.println(F("@@@@@ HTTP connect fail!"));
    return ret;
  }
  http_connect = true;
  ret = 1;
  return ret;
}

/**
 * HTTP接続設定<br>
 * @param header HTTPリクエストヘッダ
 * @return  1:成功, 0:失敗
 */
int HL7800Http::config(char* header) {
  char* rcv_buff;
  int ret =  0;
  int head_len = strlen(header);

  if (head_len < 1) {
    DBGSERIAL.print(F("@@@@@ HL7800Http::config() invalid length "));
    DBGSERIAL.println(head_len);
    return ret;
  }
  char postcmd[head_len+20];

// AT+KHTTPHEADER= <session_id> [,<local_uri>]
  sprintf_P(postcmd, PSTR("AT+KHTTPHEADER=%d\r"),1);
  DBGSERIAL.print(F("@@@@@ HL7800Http::config() cmd="));
  DBGSERIAL.println(postcmd);
  rcv_buff = serial_hl78.WriteCommandAndReadResponse(postcmd, ATRSP_CONNECT, ATRSP_ERR, MODEM_CONNECT_TIMEOUT);
  if (rcv_buff == NULL) {
    DBGSERIAL.println(F("@@@@@ HTTP POST fail!"));
    return ret;
  }
  // ヘッダ＋EOFを送る
  sprintf_P(postcmd, PSTR("%s--EOF--Pattern--"), header);
  serial_hl78.WriteCommand(postcmd);

  rcv_buff = serial_hl78.ReadResponse(ATRSP_OK, ATRSP_ERR,  MODEM_CONNECT_TIMEOUT);  
  if (rcv_buff == NULL) {
    DBGSERIAL.println(F("@@@@@ HTTP read fail!(OK not found)"));
    return ret;
  }
  
  ret = 1;
  return ret;
}

/**
 * HTTP GETを実行<br>
 * @param uri URI文字列
 * @param buff 受け取りデータバッファ
 * @param show_resp HTTP応答ヘッダを含める場合はtrue
 * @return  受信データバイト数
 */
int  HL7800Http:: get(char* uri, char* buff, bool show_resp)  {
  char* rcv_buff;
  int  ret = -1;

  // AT+KHTTPGET= <session_id>, <request_uri> [,<show_resp>]
  if (show_resp) {
    sprintf_P(cmd, PSTR("AT+KHTTPGET=1,\"%s\",1\r"),uri);
  } else {
    sprintf_P(cmd, PSTR("AT+KHTTPGET=1,\"%s\",0\r"),uri);
  }
  DBGSERIAL.print(F("@@@@@ HL7800Http::get() cmd="));
  DBGSERIAL.println(cmd);
  rcv_buff = serial_hl78.WriteCommandAndReadResponse(cmd, ATRSP_CONNECT, ATRSP_ERR, MODEM_CONNECT_TIMEOUT);
  if (rcv_buff == NULL) {
    DBGSERIAL.println(F("@@@@@ HTTP GET fail!"));
    return ret;
  }

  // データ(--EOF--Pattern--)までを受け取り
  int size = parseData(buff);
  if (size < 0) {
    return ret;
  }

  rcv_buff = serial_hl78.ReadResponse(ATRSP_OK, ATRSP_ERR,  MODEM_CONNECT_TIMEOUT);  
  if (rcv_buff == NULL) {
    DBGSERIAL.println(F("@@@@@ HTTP read fail!(OK not found)"));
    return ret;
  }

//  DBGSERIAL.println(F("@@@@@ HTTP OK!"));
  // HTTPレスポンス応答(+KHTTP_IND)を受け取ってHTTPリザルトコードを抽出
  return parseRespCode();
}

/**
 * HTTP HEADを実行(メタデータの取得)<br>
 * @param uri URI文字列
 * @param buff 受け取りデータバッファ
 * @return  受信データバイト数
 */
int HL7800Http::head(char* uri, char* buff) {
  char* rcv_buff;
  int  ret = -1;
  // AT+KHTTPHEAD=<session_id>, <request_uri>
  sprintf_P(cmd, PSTR("AT+KHTTPHEAD=1,\"%s\"\r"), uri);
  DBGSERIAL.print(F("@@@@@ HL7800Http::head() cmd="));
  DBGSERIAL.println(cmd);
  rcv_buff = serial_hl78.WriteCommandAndReadResponse(cmd, ATRSP_CONNECT, ATRSP_ERR, MODEM_CONNECT_TIMEOUT);
  if (rcv_buff == NULL) {
    DBGSERIAL.println(F("@@@@@ HTTP HEAD fail!"));
    return ret;
  }

  DBGSERIAL.println(F("@@@@@ HL7800Http::head() step2"));
  // データ(--EOF--Pattern--)までを受け取り
  int size = parseData(buff);
  if (size < 0) {
    return ret;
  }

  DBGSERIAL.println(F("@@@@@ HL7800Http::head() step3"));
  rcv_buff = serial_hl78.ReadResponse(ATRSP_OK, ATRSP_ERR,  MODEM_CONNECT_TIMEOUT);  
  if (rcv_buff == NULL) {
    DBGSERIAL.println(F("@@@@@ HTTP read fail!(OK not found)"));
    return ret;
  }
  DBGSERIAL.println(F("@@@@@ HL7800Http::head() step4"));

  // HTTPレスポンス応答(+KHTTP_IND)を受け取ってHTTPリザルトコードを抽出
  return parseRespCode();
}
/**
 * HTTP POSTを実行(URI+postデータの送信)<br>
 * @param uri URIとPOSTデータを連結した文字列
 * @param buff 受け取りデータバッファ
 * @param show_resp HTTP応答ヘッダを含める場合はtrue
 * @return  受信データバイト数
 */
int HL7800Http::post(char* uri, char* buff, bool show_resp) {
  int uri_len = strlen(uri);
  char* rcv_buff;
  int  ret = -1;

  if (uri_len < 1) {
    DBGSERIAL.print(F("@@@@@ HL7800Http::post() invalid length "));
    DBGSERIAL.println(uri_len);
    return ret;
  }
  char postcmd[uri_len+20];

// AT+KHTTPPOST=<session_id>, <local_uri>, <request_uri> [,<show_resp>]
  if (show_resp) {
    sprintf_P(postcmd, PSTR("AT+KHTTPPOST=1,,\"%s\",1\r"),uri);
  } else {
    sprintf_P(postcmd, PSTR("AT+KHTTPPOST=1,,\"%s\",0\r"),uri);
  }
//  DBGSERIAL.print(F("@@@@@ HL7800Http::post() cmd="));
//  DBGSERIAL.println(postcmd);
  rcv_buff = serial_hl78.WriteCommandAndReadResponse(postcmd, ATRSP_CONNECT, ATRSP_ERR, MODEM_CONNECT_TIMEOUT);
  if (rcv_buff == NULL) {
    DBGSERIAL.println(F("@@@@@ HTTP POST fail!"));
    return ret;
  }
  // 送信データはないのでEOFのみ送る
  sprintf(postcmd,"--EOF--Pattern--");
  serial_hl78.WriteCommand(postcmd);

  // データ(--EOF--Pattern--)までを受け取り
  int size = parseData(buff);
  if (size < 0) {
    return ret;
  }

  rcv_buff = serial_hl78.ReadResponse(ATRSP_OK, ATRSP_ERR,  MODEM_CONNECT_TIMEOUT);  
  if (rcv_buff == NULL) {
    DBGSERIAL.println(F("@@@@@ HTTP read fail!(OK not found)"));
    return ret;
  }

  // HTTPレスポンス応答(+KHTTP_IND)を受け取ってHTTPリザルトコードを抽出
  return parseRespCode();
}

/**
 * HTTP PUTを実行<br>
 * 指定したURIにデータを送信する<br>
 * @param uri URI文字列
 * @param sndbuff 送信データバッファ
 * @param rcvbuff 受け取りデータバッファ
 * @param show_resp HTTP応答ヘッダを含める場合はtrue
 * @return  受信データバイト数
 */
int HL7800Http::put(char* uri, char* sndbuff, char* rcvbuff, bool show_resp) {
  char* rcv_buff;
  int  ret = -1;

// AT+KHTTPPUT=<session_id>, <local_uri>, <request_uri> [,<show_resp>]
  if (show_resp) {
    sprintf_P(cmd, PSTR("AT+KHTTPPUT=1,,\"%s\",1\r"),uri);
  } else {
    sprintf_P(cmd, PSTR("AT+KHTTPPUT=1,,\"%s\",0\r"),uri);
  }
//  DBGSERIAL.print(F("@@@@@ HL7800Http::put() cmd="));
//  DBGSERIAL.println(cmd);
  rcv_buff = serial_hl78.WriteCommandAndReadResponse(cmd, ATRSP_CONNECT, ATRSP_ERR, MODEM_CONNECT_TIMEOUT);
  if (rcv_buff == NULL) {
    DBGSERIAL.println(F("@@@@@ HTTP POST fail!"));
    return ret;
  }

  // 送信データを送る
  serial_hl78.WriteCommand(sndbuff);
  sprintf(cmd,"--EOF--Pattern--");
  serial_hl78.WriteCommand(cmd);

  // データ(--EOF--Pattern--)までを受け取り
  int size = parseData(rcvbuff);
  if (size < 0) {
    return ret;
  }

  rcv_buff = serial_hl78.ReadResponse(ATRSP_OK, ATRSP_ERR,  MODEM_CONNECT_TIMEOUT);  
  if (rcv_buff == NULL) {
    DBGSERIAL.println(F("@@@@@ HTTP read fail!(OK not found)"));
    return ret;
  }

  // HTTPレスポンス応答(+KHTTP_IND)を受け取ってHTTPリザルトコードを抽出
  return parseRespCode();
}

/**
 * 接続状態を取得<br>
 * @param なし
 * @return  true:成功, false:失敗
 */
bool HL7800Http::connected() {
  return http_connect;
}

/**
 * HTTPセッションをクローズ<br>
 * @param なし
 * @return なし
 */
void HL7800Http::close() {
  // AT+KHTTPCLOSE= <session_id> [,<keep_cfg>]
  char* rcv_buff;
  
  if (!http_connect) {
    //  未接続の場合は何もしない
    return;
  }
  
  // AT+KHTTPCLOSE= <session_id> [,<keep_cfg>]
  sprintf_P(cmd,PSTR("AT+KHTTPCLOSE=%d\r"),1);
//  DBGSERIAL.print(F("@@@@@ HL7800Http::close() cmd="));
//  DBGSERIAL.println(cmd);
  rcv_buff = serial_hl78.WriteCommandAndReadResponse(cmd, ATRSP_OK, ATRSP_ERR, MODEM_CONNECT_TIMEOUT);

  // AT+KHTTPDEL=<session_id>
  sprintf_P(cmd,PSTR("AT+KHTTPDEL=%d\r"),1);
//  DBGSERIAL.print(F("@@@@@ HL7800Http::close() cmd="));
//  DBGSERIAL.println(cmd);
  rcv_buff = serial_hl78.WriteCommandAndReadResponse(cmd, ATRSP_OK, ATRSP_ERR, MODEM_CONNECT_TIMEOUT);
  if (rcv_buff == NULL) {
    return;
  }
//  DBGSERIAL.println(F("@@@@Success HTTP close"));
}
