/**
 * @file hl7800_modem.cpp
 * @brief HL7800 modem control API <br>
 * 
 * @date  2019/9/5 Preview-Alpha
 * @date  2019/10/3 固定文字列をプログラム領域に固定
 * @date  2019/10/10 スリープ復帰用のモデム応答待ちAPIを追加 waitModem()
 * @author kt-nakamura@kddi-tech.com
 */

#include "hl7800_modem.h"
#include "hl7800_serial.h"
#include "common.h"

PROGMEM const char ATRSP_OK[]="OK\r";
PROGMEM const char ATRSP_ERR[]="ERROR\r";
PROGMEM const char ATRSP_READY[]="READY";
PROGMEM const char ATRSP_KCNX_IND1[]="+KCNX_IND: 1,1,0";
PROGMEM const char ATRSP_KCNX_IND2[]="+KCNX_IND: 1,2";
PROGMEM const char ATRSP_KCNX_IND3[]="+KCNX_IND: 1,3\r";
PROGMEM const char ATRSP_KCNX_LF[]="\r";

HL7800Serial serial_hl78;

/**
 * モデムライブラリ
 * @param なし
 * @return なし
 */
HL7800Modem::HL7800Modem() {
  activated = false;
}

/**
 * モデム初期化<br>
 * モデムへのHWリセットから初期パラメータ設定までを行う<br>
 * 
 * @param なし
 * @return  true:成功, false:失敗
 */
bool HL7800Modem::Init(void) {
  serial_hl78.Init();
  char cmd[40];
  char* rcv_buff;
//  int rcvlen;

//  DBGSERIAL.println(F("@@@@@ HL7800Modem::Init()"));
  DBGSERIAL.println(F("Modem Init"));
  activated = false;

  // Modem PowerON and WAKEUP
  pinMode(MDMWAKEUP , OUTPUT);
  digitalWrite(MDMWAKEUP, HIGH);
  pinMode(MDMPWRON , INPUT); // ドライブしない
  pinMode(MDMDTR , OUTPUT);
  digitalWrite(MDMDTR, LOW);
  pinMode(MDMCTS , INPUT); // 入力
  pinMode(MDMRI , INPUT); // 入力

  // HW reset
  pinMode(MDMRST , OUTPUT);
  digitalWrite(MDMRST, LOW);
  delay(100);
  digitalWrite(MDMRST, HIGH);

  // モデム起動待ち
  if (!waitModem() ) {
    return false;
  }
  DBGSERIAL.println(F("start modem"));

  // エラーメッセージのフォーマット設定
  sprintf_P(cmd,PSTR("AT+CMEE=%d\r"),1);
  rcv_buff = serial_hl78.WriteCommandAndReadResponse(cmd, ATRSP_OK, ATRSP_ERR, MODEM_SHORT_TIMEOUT);
  if (rcv_buff == NULL) {
    return false;
  }

  // スリープ禁止
  sprintf_P(cmd,PSTR("AT+KSLEEP=%d\r"),2);
  rcv_buff = serial_hl78.WriteCommandAndReadResponse(cmd, ATRSP_OK, ATRSP_ERR, MODEM_SHORT_TIMEOUT);
  if (rcv_buff == NULL) {
    return false;
  }
  digitalWrite(MDMWAKEUP, LOW); // WAKEUPを解除する

  // use Hardware-Flow
  sprintf_P(cmd,PSTR("AT&K%d\r"),3);
  rcv_buff = serial_hl78.WriteCommandAndReadResponse(cmd, ATRSP_OK, ATRSP_ERR, MODEM_SHORT_TIMEOUT);
  if (rcv_buff == NULL) {
    return false;
  }

  // LTE band set
  sprintf_P(cmd,PSTR("AT+KBNDCFG=%d,00000000000000020000\r"),0);
  rcv_buff = serial_hl78.WriteCommandAndReadResponse(cmd, ATRSP_OK, ATRSP_ERR, MODEM_SHORT_TIMEOUT);
  if (rcv_buff == NULL) {
    return false;
  }

  // set Network Registration Status
  sprintf_P(cmd,PSTR("AT+CEREG=%d\r"),0);
  rcv_buff = serial_hl78.WriteCommandAndReadResponse(cmd, ATRSP_OK, ATRSP_ERR, MODEM_SHORT_TIMEOUT);
  if (rcv_buff == NULL) {
    return false;
  }

  // set Report Mobile Termination Error
  sprintf_P(cmd,PSTR("AT+CMEE=%d\r"),0);
  rcv_buff = serial_hl78.WriteCommandAndReadResponse(cmd, ATRSP_OK, ATRSP_ERR, MODEM_SHORT_TIMEOUT);
  if (rcv_buff == NULL) {
    return false;
  }

  // 初期化成功
  return true;
}

/**
 * PDP接続を行う<br>
 * @param accessPointName APN名
 * @param userName 接続ユーザー名
 * @param password 接続パスワード
 * @param waitForRegistTimeoutMs 接続までのタイムアウト(ms)
 * @return  true:成功, false:失敗
 */
bool HL7800Modem::activatePdp(const char* accessPointName, const char* userName, const char* password, long waitForRegistTimeoutMs) {
  char cmd[100];
//  int rcvlen;
  char* rcv_buff;
//  DBGSERIAL.println(F("@@@@@ HL7800Modem::activatePdp()"));
  if (activated) {
    DBGSERIAL.println(F("ERROR: already activated!"));
    return false;
  }

  // APN設定前にPDU接続を解放
  sprintf_P(cmd,PSTR("AT+CGATT=%d\r"),0);
  rcv_buff = serial_hl78.WriteCommandAndReadResponse(cmd, ATRSP_OK, ATRSP_ERR, MODEM_DISCONNECT_TIMEOUT);
  if (rcv_buff == NULL) {
    return false;
  }

  // モデムにAPN登録 PROFILE 1
  sprintf_P(cmd,PSTR("AT+KCNXCFG=1\,\"GPRS\"\,\"%s\"\,\"%s\"\,\"%s\"\,\"IPV4\"\,\"0.0.0.0\"\,\"0.0.0.0\"\,\"0.0.0.0\"\r"),accessPointName, userName, password);
//  DBGSERIAL.print(F("@@@@@ HL7800Modem::activatePdp() cmd= "));
//  DBGSERIAL.println(cmd);

  rcv_buff = serial_hl78.WriteCommandAndReadResponse(cmd, ATRSP_OK, ATRSP_ERR, MODEM_SHORT_TIMEOUT);
  if (rcv_buff == NULL) {
    return false;
  }
//  DBGSERIAL.println(F("@@@@OK"));
//  DBGSERIAL.println(rcv_buf);

  // PDP接続にPROFILE 1を指定
  sprintf_P(cmd,PSTR("AT+KCNXPROFILE=%d\r"),1);
//  DBGSERIAL.print(F("@@@@@ HL7800Modem::activatePdp() cmd= "));
//  DBGSERIAL.println(cmd);
  rcv_buff = serial_hl78.WriteCommandAndReadResponse(cmd, ATRSP_OK, ATRSP_ERR, MODEM_SHORT_TIMEOUT);
  if (rcv_buff == NULL) {
    return false;
  }
//  DBGSERIAL.println(F("@@@@OK"));

  sprintf_P(cmd,PSTR("AT+WPPP=2\,1\,\"%s\"\,\"%s\"\r"),userName, password);
//  DBGSERIAL.print(F("@@@@@ HL7800Modem::activatePdp() cmd= "));
//  DBGSERIAL.println(cmd);
  rcv_buff = serial_hl78.WriteCommandAndReadResponse(cmd, ATRSP_OK, ATRSP_ERR, MODEM_SHORT_TIMEOUT);
  if (rcv_buff == NULL) {
    return false;
  }
//  DBGSERIAL.println(F("@@@@OK"));
//  DBGSERIAL.println(rcv_buf);

  // PDPを有効
  sprintf_P(cmd,PSTR("AT+KCNXUP=%d\r"),1);
//  DBGSERIAL.print(F("@@@@@ HL7800Modem::activatePdp() cmd= "));
//  DBGSERIAL.println(cmd);
  rcv_buff = serial_hl78.WriteCommandAndReadResponse(cmd, ATRSP_OK, ATRSP_ERR, waitForRegistTimeoutMs);
  if (rcv_buff == NULL) {
    return false;
  }
//  DBGSERIAL.println(rcv_buf);
//  DBGSERIAL.println(F("@@@@OK"));

  rcv_buff = serial_hl78.ReadResponse(ATRSP_KCNX_IND1, ATRSP_KCNX_IND2,  waitForRegistTimeoutMs);  
  if (rcv_buff == NULL) {
    // 接続失敗の場合は受信バッファのゴミを捨てる
    DBGSERIAL.println(F("Failed to PDP connect!"));
    rcv_buff = serial_hl78.ReadResponse(ATRSP_KCNX_LF, ATRSP_ERR,  MODEM_SHORT_TIMEOUT);  
    return false;
  }
//  DBGSERIAL.println(F("@@@@Success connect"));
  
  activated = true;
  return true;
}

/**
 * PDP切断<br>
 * 
 * @param なし 
 * @return  true:成功, false:失敗
 */
bool HL7800Modem::deactivate(void) {
//  DBGSERIAL.print(F("@@@@@ HL7800Modem::deactivate() "));
  if (activated) {
    char cmd[20];
//    int rcvlen;
    char* rcv_buff;
    DBGSERIAL.println(F("exec deactivate"));
    sprintf_P(cmd,PSTR("AT+KCNXDOWN=1,%d\r"),1);  // 有効なセッションが残っていても切断する
    rcv_buff = serial_hl78.WriteCommandAndReadResponse(cmd, ATRSP_OK, ATRSP_ERR, MODEM_SHORT_TIMEOUT);
    if (rcv_buff == NULL) {
      return false;
    }

    rcv_buff = serial_hl78.ReadResponse(ATRSP_KCNX_IND3, ATRSP_ERR,  MODEM_DISCONNECT_TIMEOUT);  
    if (rcv_buff == NULL) {
      DBGSERIAL.println(F("@@@@@ PDP disconnect timeout!"));
      return false;
    }
    activated = false;
  }
  DBGSERIAL.println(F("PDP disconnected"));
  digitalWrite(MDMDTR, LOW);
  return true;
}

/**
 * PDP接続チェック<br>
 * 
 * @param なし 
 * @return  true:接続中, false:接続なし
 */
bool HL7800Modem::isActive(void) {
  return activated;
}

/**
 * モデム応答チェック<br>
 * 
 * @param なし 
 * @return  true:成功, false:モデム応答なし
 */
bool HL7800Modem::chkModem(void) {
  char cmd[10];
  char* rcv_buff;

  sprintf_P(cmd,PSTR("AT%s\r"),"");
  rcv_buff = serial_hl78.WriteCommandAndReadResponse(cmd, ATRSP_OK, ATRSP_ERR, MODEM_SHORT_TIMEOUT);
  if (rcv_buff == NULL) {
    return false;
  }
  return true;
}

/**
 * モデム応答待ち<br>
 * スリープ復帰等でATコマンド応答可能となるまで時間がある場合があるので、このAPIで応答待ちを行う
 * @param なし 
 * @return  true:成功, false:モデム応答なし
 */
bool HL7800Modem::waitModem(void) {
  char cmd[20];
  char* rcv_buff;

    // モデム起動待ち
#ifdef USE_SOFTSERIAL
  sprintf_P(cmd,PSTR("ATE0\r"));  // SoftSerialは送受信を同時にできないためエコーバックを抑制
#else
  sprintf_P(cmd,PSTR("ATE%d\r"),1);  // HardSerialは送受信同時にできるためコマンド可視化のためエコーバックを入れる
#endif
  int loop=0;
  for (loop=0;loop<10;loop++) {
//  DBGSERIAL.print(F("@@@@@ HL7800Modem::waitModem() cmd= "));
//  DBGSERIAL.println(cmd);

    rcv_buff = serial_hl78.WriteCommandAndReadResponse(cmd, ATRSP_OK, ATRSP_ERR, MODEM_SHORT_TIMEOUT);
    if (rcv_buff != NULL) {
      break;
    }
    delay(1000);
  }
  if (loop > 9) {
    return false;
  }
  return true;
}



/**
 * データ配列から指定した文字に挟まれた文字列を取得する<br>
 * 入力されたデータ配列にNULL終端を上書きするので注意<br>
 * @param str 判定したいデータ配列
 * @param pattrnStart 開始パターン文字列
 * @param pattrnEnd 終端パターン文字列
 * @return 判定後の文字列(NULL終端)
 */
char* serchResp(char* str, const char* pattrnStart, const char* pattrnEnd) {
  char* head = strstr(str,pattrnStart);
  if (head == NULL) {
    return NULL;
  }
  char* end = strstr(head + strlen(pattrnStart),pattrnEnd);
  if (end == NULL) {
    return NULL;
  }
  *end = NULL;
  return head + strlen(pattrnStart); // 検索文字分オフセットを掛けて返す
}

/**
 * データ配列に含まれるIPアドレス文字を変換<br>
 * @param strIp 判定したいデータ配列
 * @return IPアドレス
 */
IPAddress parseIp(char* strIp) {
  char* strInput = strIp;
  char* next;
  IPAddress resp = {0,0,0,0};
  int strVal[4];

//  DBGSERIAL.print("parseIp() ");
  for (int i=0;i<3;i++) {
    next = strstr(strInput,".");
    if (next == NULL) return resp;
    *next = NULL;
    strVal[i] = atoi(strInput);
//    DBGSERIAL.print(strVal[i]);
//    DBGSERIAL.print(".");
    strInput = next+1;
  }
  strVal[3] = atoi(strInput);
//  DBGSERIAL.print(strVal[3]);
//  DBGSERIAL.println("");

  resp[0] = strVal[0];
  resp[1] = strVal[1];
  resp[2] = strVal[2];
  resp[3] = strVal[3];
  return resp;
}

/**
 * PDP接続中の WWAN IPアドレスを取得<br>
 * @param なし
 * @return IPアドレス
 */
IPAddress HL7800Modem::getIpAddress(void) {
  char cmd[16];
//  int rcvlen;
  char* rcv_buff;
  IPAddress resp = {0,0,0,0};

  sprintf_P(cmd,PSTR("AT+KCGPADDR%s\r"),"");
//  DBGSERIAL.print(F("@@@@@ HL7800Modem::getIpAddress() cmd= "));
//  DBGSERIAL.println(cmd);
  rcv_buff = serial_hl78.WriteCommandAndReadResponse(cmd, ATRSP_OK, ATRSP_ERR, MODEM_SHORT_TIMEOUT);
  if (rcv_buff == NULL) {
    return resp;
  }
  
//  DBGSERIAL.println(F("@@@@OK"));
//  DBGSERIAL.println(rcv_buff);
  char* head= serchResp(rcv_buff, "+KCGPADDR: 1,\"","\"");
  if (head != NULL) {
//    DBGSERIAL.print(F("@@@@@ resp= "));
//    DBGSERIAL.println(head);
    resp = parseIp(head);
  }
  return resp;
}

/**
 * PDP接続中の WWAN IPアドレスから接続判定<br>
 * @param なし
 * @return  1:有効IPアドレス, 0:無効IPアドレス
 */
bool HL7800Modem::chkIpAddress(void) {
  IPAddress ip = getIpAddress();
  DBGSERIAL.print(F("WWAN IP=["));
  DBGSERIAL.print(ip[0]);  DBGSERIAL.print(",");
  DBGSERIAL.print(ip[1]);  DBGSERIAL.print(",");
  DBGSERIAL.print(ip[2]);  DBGSERIAL.print(",");
  DBGSERIAL.print(ip[3]);
  DBGSERIAL.println("]");

  if ((ip[0] | ip[1] | ip[2] | ip[3]) == 0) {
    // IPアドレス無しの場合は不正とする
    return false;
  }
  return true;
}

/**
 * 電界強度(RSSI)を取得<br>
 * 戻り値は１～４０で９９は圏外<br>
 * @param なし
 * @return RSSI値
 */
int HL7800Modem::getRssi(void) {
  // at+csq
  // +CSQ: 18,99
  char cmd[16];
  char* rcv_buff;

  sprintf_P(cmd,PSTR("AT+CSQ%s\r"),"");
//  DBGSERIAL.print("@@@@@ HL7800Modem::getRssi() cmd= ");
//  DBGSERIAL.println(cmd);
  rcv_buff = serial_hl78.WriteCommandAndReadResponse(cmd, ATRSP_OK, ATRSP_ERR, MODEM_SHORT_TIMEOUT);
  if (rcv_buff == NULL) {
    return -1;
  }
//  DBGSERIAL.println(F("@@@@OK"));
  char* head = serchResp(rcv_buff, "+CSQ:",",");
  if (head != NULL) {
//    DBGSERIAL.print("@@@@@ resp= ");
//    DBGSERIAL.println(head);
    return atoi(head);
  }
  return -1;
}

/**
 * SIMのPINチェック<br>
 * @param なし
 * @return  1:成功, 0:失敗
 */
bool HL7800Modem::chkPin(void) {
  char cmd[16];
  char* rcv_buff;

  sprintf_P(cmd,PSTR("AT+CPIN?%s\r"),"");
//  DBGSERIAL.print("@@@@@ HL7800Modem::chkPin() cmd= ");
//  DBGSERIAL.println(cmd);
  rcv_buff = serial_hl78.WriteCommandAndReadResponse(cmd, ATRSP_OK, ATRSP_ERR, MODEM_SHORT_TIMEOUT);
  if (rcv_buff == NULL) {
    return false;
  }

  // 応答にREADYがあるか？
  char* ready = strstr_P(rcv_buff,ATRSP_READY);
  if (ready == NULL) {
    return false;
  }
  return true;
}

/**
 * モデムのIMEI取得<br>
 * @param buff 受け取りデータバッファ
 * @return  IMEIの文字列長
 */
int HL7800Modem::getImei(char* buff) {
  int  ret = -1;
  char cmd[16];
  char* rcv_buff;

  sprintf_P(cmd,PSTR("AT+CGSN=%d\r"),1);
//  DBGSERIAL.print("@@@@@ HL7800Modem::getImei() cmd= ");
//  DBGSERIAL.println(cmd);
  rcv_buff = serial_hl78.WriteCommandAndReadResponse(cmd, ATRSP_OK, ATRSP_ERR, MODEM_SHORT_TIMEOUT);
  if (rcv_buff == NULL) {
    return ret;
  }

  char* head= serchResp(rcv_buff, "+CGSN:\"" , "\"");
  if (head != NULL) {
//    DBGSERIAL.print("@@@@@ HL7800Modem::getImei() resp=");
//    DBGSERIAL.println(head);
    int len = strlen(head);
    memcpy(buff,head,len);
    return len;
  }

  return ret;
}

/**
 * モデムのCCID取得<br>
 * @param buff 受け取りデータバッファ
 * @return  CCIDの文字列長
 */
int HL7800Modem::getCcid(char* buff) {
  int  ret = -1;
  char cmd[16];
  char* rcv_buff;

  sprintf_P(cmd,PSTR("AT+CCID?%s\r"),"");
//  DBGSERIAL.print("@@@@@ HL7800Modem::getCcid() cmd= ");
//  DBGSERIAL.println(cmd);
  rcv_buff = serial_hl78.WriteCommandAndReadResponse(cmd, ATRSP_OK, ATRSP_ERR, MODEM_SHORT_TIMEOUT);
  if (rcv_buff == NULL) {
    return ret;
  }

  char* head= serchResp(rcv_buff, "+CCID: " , "\r");
  if (head != NULL) {
//    DBGSERIAL.print("@@@@@ HL7800Modem::getCcid() resp=");
//    DBGSERIAL.println(head);
    int len = strlen(head);
    memcpy(buff,head,len);
    return len;
  }
  return ret;
}

/**
 * モデム内時間取得<br>
 * 成功時は次のような文字列が取得できる<br>
 * "yy/MM/dd,hh:mm:ss±zz"<br>
 * @param buff 受け取りデータバッファ
 * @return  時計の文字列長
 */
int HL7800Modem::getRTC(char* buff) {
  int  ret = -1;
  char cmd[16];
  char* rcv_buff;

  // AT+CCLK?
  sprintf_P(cmd,PSTR("AT+CCLK?%s\r"),"");
//  DBGSERIAL.print("@@@@@ HL7800Modem::getRTC() cmd= ");
//  DBGSERIAL.println(cmd);
  rcv_buff = serial_hl78.WriteCommandAndReadResponse(cmd, ATRSP_OK, ATRSP_ERR, MODEM_SHORT_TIMEOUT);
  if (rcv_buff == NULL) {
    return ret;
  }

  char* head= serchResp(rcv_buff, "+CCLK: " , "\r");
  if (head != NULL) {
//    DBGSERIAL.print("@@@@@ HL7800Modem::getRTC() resp=");
//    DBGSERIAL.println(head);
    int len = strlen(head);
    memcpy(buff,head,len);
    return len;
  }
  return ret;
}

/**
 * モデムeDRX設定<br>
 * モデムにeDRXパラメータを設定する<br>
 * @param edrx_value eDRX設定値<br>
 * String type; half a byte in a 4-bit format. The eDRX value refers to bit 4 to 1 of octet 3 of the Extended DRX parameters information element.<br>
 * 設定値が-1の場合はeDRX無効とする<br>
 * @return  true:成功, false:モデム応答なし
 */
bool HL7800Modem::setEDRX(int edrx_value) {
  int  ret = false;
  char cmd[16];
  char* rcv_buff;

  // AT+CEDRXS= [<mode>[,<AcT-type> [,<Requested_ eDRX_value>]]]
  if (edrx_value < 0) {
    sprintf_P(cmd,PSTR("AT+CEDRXS=%d\r"),0);
  } else {
    sprintf_P(cmd,PSTR("AT+CEDRXS=%d,4,%d\r"),1,edrx_value);
  }
  DBGSERIAL.print("@@@@@ HL7800Modem::setEDRX() cmd= ");
  DBGSERIAL.println(cmd);
  rcv_buff = serial_hl78.WriteCommandAndReadResponse(cmd, ATRSP_OK, ATRSP_ERR, MODEM_SHORT_TIMEOUT);
  if (rcv_buff == NULL) {
    return ret;
  }

  return true;
}

/**
 * モデム省電力設定<br>
 * モデムを省電力状態に移行する<br>
 * @param mode 省電力モード 0:SLEEP_MODE 1:LIGHT_HIBERNATE 2:HIBERNATE
 * @return  true:成功, false:モデム応答なし
 */
bool HL7800Modem::enterSleep(int mode) {
  int  ret = false;
  char cmd[16];
  char* rcv_buff;

  // AT+KSLEEP= <mngt>[,<level> [,<delay>]]
  // mngt=1 Standalone sleep mode. The module decides by itself when it enters sleep mode.
  sprintf_P(cmd,PSTR("AT+KSLEEP=1,%d\r"),mode);
//  sprintf_P(cmd,PSTR("AT+KSLEEP=1,%d,1\r"),mode);
//  DBGSERIAL.print("@@@@@ HL7800Modem::leaveSleep() cmd= ");
//  DBGSERIAL.println(cmd);
  rcv_buff = serial_hl78.WriteCommandAndReadResponse(cmd, ATRSP_OK, ATRSP_ERR, MODEM_SHORT_TIMEOUT);
  if (rcv_buff == NULL) {
    return ret;
  }
  digitalWrite(MDMWAKEUP, LOW); // WAKEUP信号を止める
//  delay(10000);

  // HIBERNATEではVGPIOの出力が止まることからモデム入力信号から逆流しないようLowレベルに落とす
  if (mode == HIBERNATE) {
    digitalWrite(MDMDTR, LOW);
    digitalWrite(MDMRTS, LOW);
    digitalWrite(MDMTX,LOW); 
    pinMode(MDMTX , OUTPUT);
  }
  ret = true;
  return ret;
}

/**
 * モデム省電力復帰<br>
 * 省電力状態からの復帰はWAKEUP信号を出してからAT+KSLEEP=2で省電力設定を解除することにより実現する<br>
 * @param なし
 * @return  true:成功, false:モデム応答なし
 */
bool  HL7800Modem::leaveSleep() {
  int  ret = false;
  char cmd[16];
  char* rcv_buff;

  serial_hl78.Init(); // シリアルポートを再設定

  digitalWrite(MDMWAKEUP, HIGH); // WAKEUP信号を出す
  // モデム起動待ち
  if (!waitModem() ) {
    return false;
  }

  // スリープ設定を解除する
  sprintf_P(cmd,PSTR("AT+KSLEEP=%d\r"),2);
//  DBGSERIAL.print(F("@@@@@ HL7800Modem::leaveSleep() cmd= "));
//  DBGSERIAL.println(cmd);
  rcv_buff = serial_hl78.WriteCommandAndReadResponse(cmd, ATRSP_OK, ATRSP_ERR, MODEM_SHORT_TIMEOUT);
  if (rcv_buff == NULL) {
    return ret;
  }
  digitalWrite(MDMWAKEUP, LOW); // WAKEUP信号を止める
//  digitalWrite(MDMDTR, HIGH);
  delay(1000);
  ret = true;
  return ret;
}

/**
 * モデム電源オフ<br>
 * モデムを電源オフに移行する
 * 復帰はPOWER_ON_N信号で行う
 * @param fast 
 * @return  true:成功, false:モデム応答なし
 */
bool HL7800Modem::powerDown(void) {
  int  ret = false;
  char cmd[20];
  char* rcv_buff;

  // 復帰条件となるPWR_ON_N 入力を有効にする
  sprintf_P(cmd,PSTR("AT+KHWIOCFG=0,%d\r"),1);
//  DBGSERIAL.print(F("@@@@@ HL7800Modem::powerDown() cmd= "));
//  DBGSERIAL.println(cmd);
  rcv_buff = serial_hl78.WriteCommandAndReadResponse(cmd, ATRSP_OK, ATRSP_ERR, MODEM_SHORT_TIMEOUT);
  if (rcv_buff == NULL) {
    return ret;
  }

  sprintf_P(cmd,PSTR("AT+CPWROFF=%d\r"),1);
//  DBGSERIAL.print(F("@@@@@ HL7800Modem::powerDown() cmd= "));
//  DBGSERIAL.println(cmd);
  rcv_buff = serial_hl78.WriteCommandAndReadResponse(cmd, ATRSP_OK, ATRSP_ERR, MODEM_SHORT_TIMEOUT);
  if (rcv_buff == NULL) {
    return ret;
  }

  digitalWrite(MDMDTR, LOW);
  digitalWrite(MDMRTS, LOW);
  digitalWrite(MDMTX,LOW); 
  pinMode(MDMTX , OUTPUT);

  ret = true;
  return ret;
}

/**
 * モデム電源復帰<br>
 * PowerDown状態となったモデムを復帰させる<br>
 * 復帰時は電源ONと同じであるため使用前に初期化を行うこと
 * @param なし
 * @return  true:成功, false:モデム応答なし
 */
bool  HL7800Modem::powerUp() {
  pinMode(MDMPWRON , OUTPUT);
  digitalWrite(MDMPWRON, LOW);
  delay(1000);
  digitalWrite(MDMPWRON, HIGH);
  pinMode(MDMPWRON , INPUT); // ドライブしない
  return true;
}
