/**
 * @file hl7800_serial.cpp
 * @brief HL7800 serial driver <br>
 * 
 * @date  2019/9/5 Preview-Alpha
 * @author kt-nakamura@kddi-tech.com
 */

#include "hl7800_serial.h"
#include "common.h"

char rxSerBuff[RCVBUFF_SIZE];

#ifdef USE_SOFTSERIAL
#include "SoftwareSerial.h"
SoftwareSerial MDMSERIAL(MDMRX, MDMTX); // RX, TX
#endif
#ifdef USE_ALTSERIAL
#include "wiring_private.h"
#endif

/**
 * シリアルドライバ初期化<br>
 * @param なし
 * @return なし
 */
void HL7800Serial::Init(void) {
  pinMode(MDMRTS , OUTPUT);
  digitalWrite(MDMRTS, LOW);
  MDMSERIAL.begin(MDMBPS);
#ifdef USE_ALTSERIAL
  pinPeripheral(MDMTX, PIO_SERCOM_ALT);
  pinPeripheral(MDMRX, PIO_SERCOM_ALT);
#endif
}

/**
 * モデムにコマンド/データ送信<br>
 * @param command 送信したいバイト列(NULL終端)
 * @return なし
 */
void HL7800Serial::WriteCommand(char* command) {
  char* xmit_p = command;
  int xmitlen = strlen(command);

//  DBGSERIAL.print(F("@@@@@ HL7800Serial::WriteCommand() len="));
//  DBGSERIAL.println(xmitlen);
  for(;;) {
    // Transmit Command
    if (xmitlen > 0) {
      MDMSERIAL.write(*xmit_p++);
      xmitlen--;
    } else {
      // transmit complete
      break;
    }
  }
  MDMSERIAL.flush();
}

/**
 * モデムにバイナリデータ送信<br>
 * @param data　データバイト列 
 * @param length データ長
 * @return なし
 */
void HL7800Serial::WriteData(const uint8_t* data, int length) {
  uint8_t* xmit_p = (uint8_t*)data;
  int xmitlen = length;

//  DBGSERIAL.print(F("@@@@@ HL7800Serial::WriteData() len="));
//  DBGSERIAL.println(xmitlen);
  for(;;) {
    // Transmit Command
    if (xmitlen > 0) {
      MDMSERIAL.write(*xmit_p++);
      xmitlen--;
    } else {
      // transmit complete
      break;
    }
  }
  MDMSERIAL.flush();
}

/**
 * モデム受信待ち<br>
 * @param patternOK 期待する成功文字列
 * @param patternNG 期待するエラー文字列
 * @param timeout_ms タイムアウト時間(ms)
 * @return  受信データ(NULL終端)
 */
char* HL7800Serial::ReadResponse(const char* patternOK, const char* patternNG, unsigned long timeout_ms) {
  char* rcv_p = rxSerBuff;
  int rcvlen = 0;

  digitalWrite(MDMRTS, HIGH);
  memset(rxSerBuff,0x0,RCVBUFF_SIZE);
  unsigned long currentTime;
  unsigned long endTime = millis() + timeout_ms;

  digitalWrite(MDMRTS, LOW);
  for(;;) {
    if (MDMSERIAL.available() > 0) {
      digitalWrite(MDMRTS, HIGH);
      char rcv = MDMSERIAL.read();
      rcvlen++;
      if (rcvlen > RCVBUFF_SIZE) {
        DBGSERIAL.println(F("recive data is too big!"));
        return NULL;
      }
      *rcv_p++ = rcv; 
      if (strstr_P(rxSerBuff,patternOK) != NULL) {
//        DBGSERIAL.println(F("@@@@ responseOK!"));
        *rcv_p++ = 0; 
        return rxSerBuff;
      }
      if (strstr_P(rxSerBuff,patternNG) != NULL) {
//        DBGSERIAL.println(F("@@@@ responseNG!"));
        return NULL;
      }
      
    } else {
        digitalWrite(MDMRTS, LOW);
    }
    
    currentTime = millis();
    if (endTime < currentTime) {
//      DBGSERIAL.println(retval);
      DBGSERIAL.println(F("command timeout!"));
      digitalWrite(MDMRTS, LOW);
      return NULL;
    }
  }
}
/**
 * モデムへコマンド送信とレスポンス受信を同時に行う<br>
 * @param command 送信したいバイト列(NULL終端)
 * @param patternOK 期待する成功文字列
 * @param patternNG 期待するエラー文字列
 * @param timeout_ms タイムアウト時間(ms)
 * @return  受信データ(NULL終端)
 */
char* HL7800Serial::WriteCommandAndReadResponse(char* command, const char* patternOK, const char* patternNG, unsigned long timeout_ms) {
  char* xmit_p = command;
  char* rcv_p = rxSerBuff;
  int xmitlen = strlen(command);
  int rcvlen = 0;

//  DBGSERIAL.print(F("@@@@@ HL7800Serial::WriteCommandAndReadResponse() len="));
//  DBGSERIAL.println(xmitlen);
  memset(rxSerBuff,0x0,RCVBUFF_SIZE);

  unsigned long currentTime;
  unsigned long endTime = millis() + timeout_ms;

  digitalWrite(MDMRTS, LOW);
  for(;;) {
    // Transmit Command
    if (xmitlen > 0) {
      MDMSERIAL.write(*xmit_p++);
      xmitlen--;
    }

    if (MDMSERIAL.available() > 0) {
      digitalWrite(MDMRTS, HIGH);
      char rcv = MDMSERIAL.read();
      rcvlen++;
      if (rcvlen > RCVBUFF_SIZE) {
        DBGSERIAL.println(F("recive data is too big!"));
        return NULL;
      }
      *rcv_p++ = rcv; 
     
//      DBGSERIAL.write(rcv);
      if (strstr_P(rxSerBuff,patternOK) != NULL) {
//        DBGSERIAL.println(F("@@@@ responseOK!"));
        *rcv_p++ = 0;
        digitalWrite(MDMRTS, LOW);
        return rxSerBuff;
      }
      if (strstr_P(rxSerBuff,patternNG) != NULL) {
        DBGSERIAL.println(F("@@@@ responseNG!"));
        digitalWrite(MDMRTS, LOW);
        return NULL;
      }
    } else {
        digitalWrite(MDMRTS, LOW);
    }
    
    currentTime = millis();
    if (endTime < currentTime) {
//      DBGSERIAL.println(retval);
      DBGSERIAL.println(F("command timeout!"));
      digitalWrite(MDMRTS, LOW);
      return NULL;
    }
  }
  return NULL;
}
