/**
 * @file hl7800_udp.cpp
 * @brief  HL7800 modem UDP client API <br>
 * 
 * @date  2019/9/5 Preview-Alpha
 * @date  2019/9/27 バイナリ送信への対処
 * @author kt-nakamura@kddi-tech.com
 */

#include "common.h"
#include "hl7800_udp.h"
#include "hl7800_serial.h"
#include "hl7800_modem.h"
#include <Udp.h>

PROGMEM const char ATRSP_OK[]="OK\r";
PROGMEM const char ATRSP_ERR[]="ERROR\r";
PROGMEM const char ATRSP_KCNX_LF[]="\r";
PROGMEM const char ATDATA_EOF[]="--EOF--Pattern--";
PROGMEM const char ATRSP_CONNECT[]="CONNECT\r";
PROGMEM const char ATRSP_KUDP_IND1[]="+KUDP_IND: 1,1";
PROGMEM const char ATRSP_KUDP_IND0[]="+KUDP_IND: 1,0";

#define MODE_CLIENT 0
#define MODE_SERVER 1

char rxUdpBuff[SNDBUFF_SIZE];
int txBuffSize = 0;

extern HL7800Serial serial_hl78;

/**
 * UDPライブラリ初期化<br>
 * Initializes the UDP library and network settings. <br>
 * @param port 自デバイスのUDPポート番号
 * @return  1:成功, 0:失敗
 */
uint8_t HL7800Udp::begin(uint16_t port) {
  int ret = 0;
  char* rcv_buff;

  memcpy(_targetHost,0x0,MAX_HOSTNAME_SIZE);
  _targetIP = {0,0,0,0};
  _targetPort = 0;
  _validHost = false;
  _validIP = false;

  DBGSERIAL.print(F("@@@@@ HL7800Udp::begin(port = "));
  DBGSERIAL.print(port);
  DBGSERIAL.println(")");

// AT+KUDPCFG= [<cnx cnf>], <mode>[,[<port>][,[<data_mode>] [,[<udp remote address>][,[<udp_port>][,[<af>][, [<restore_on_ boot>]]]]]]]
  sprintf_P(cmd,PSTR("AT+KUDPCFG=1,%d,%d\r"), MODE_CLIENT, port);
//  DBGSERIAL.print(F("@@@@@ HL7800Udp::begin() cmd="));
//  DBGSERIAL.println(cmd);

  rcv_buff = serial_hl78.WriteCommandAndReadResponse(cmd, ATRSP_OK, ATRSP_ERR, MODEM_CONNECT_TIMEOUT);
  if (rcv_buff == NULL) {
    DBGSERIAL.println(F("@@@@@ HL7800Udp::begin() fail!"));
    return ret;
  }

  rcv_buff = serial_hl78.ReadResponse(ATRSP_KUDP_IND1, ATRSP_KUDP_IND0,  MODEM_LONG_TIMEOUT);
  if (rcv_buff == NULL) {
    DBGSERIAL.println(F("@@@@@ HL7800Udp::begin() config timeout!"));
    return ret;
  }
  
  DBGSERIAL.println(F("@@@@@ HL7800Udp::begin() sucess."));
  ret = 1;
  return ret;
}

/**
 * UDPセッションを破棄する<br>
 * Disconnect from the server. Release any resource being used during the UDP session. <br>
 * @param なし 
 * @return  なし
 */
void HL7800Udp::stop() {
  char* rcv_buff;
  // AT+KUDPDEL=<session_id>
  sprintf_P(cmd,PSTR("AT+KUDPDEL=%d\r"),1);

  _validHost = false;
  _validIP = false;

//  DBGSERIAL.print(F("@@@@@ HL7800Udp::stop() cmd="));
//  DBGSERIAL.println(cmd);

  rcv_buff = serial_hl78.WriteCommandAndReadResponse(cmd, ATRSP_OK, ATRSP_ERR, MODEM_CONNECT_TIMEOUT);
  if (rcv_buff == NULL) {
//    DBGSERIAL.println(F("@@@@@ HL7800Udp::stop() fail!"));
  }
}

/**
 * UDP送信開始(IPアドレス指定)<br>
 * Starts a connection to write UDP data to the remote connection<br>
 * ここでは内部変数にパラメータをコピーするのみ、実際の動作はwrite()で開始される<br>
 * @param ip 送信相手のIP
 * @param port 送信相手のUDPポート番号
 * @return  1:成功, 0:失敗
 */
int HL7800Udp::beginPacket(IPAddress ip, uint16_t port) {
  int ret = 1;
//  DBGSERIAL.println(F("@@@@@ HL7800Udp::beginPacket(IP)"));

  _validHost = false;
  _validIP = true;
  _targetIP = ip;
  _targetPort = port;
  return ret;
}

/**
 * UDP送信開始(host名指定)<br>
 * Starts a connection to write UDP data to the remote connection<br>
 * ここでは内部変数にパラメータをコピーするのみ、実際の動作はwrite()で開始される<br>
 * @param host 送信相手の
 * @param port 送信相手のUDPポート番号
 * @return  1:成功, 0:失敗
 */
int HL7800Udp::beginPacket(const char *host, uint16_t port) {
  int ret = 0;
  int len = strlen(host);
  if (( len < 1) || (len > MAX_HOSTNAME_SIZE)) {
    DBGSERIAL.println(F("host name invalid or to long!"));
    return ret;
  }
  DBGSERIAL.print(F("@@@@@ HL7800Udp::beginPacket() host="));
  DBGSERIAL.print(host);
  DBGSERIAL.print(F(" port ="));
  DBGSERIAL.println(port);
  memcpy(_targetHost,host,len);
  _targetPort = port;
  _validHost = true;
  _validIP = false;
  ret = 1;
  return ret;
}

/**
 * UDP送信終了<br>
 * Called after writing UDP data to the remote connection. <br>
 * UDP送信後の終了処理を行う<br>
 * @param なし 
 * @return  1:成功, 0:失敗
 */
//PROGMEM const char ATCMD_KUDPCLOSE[]="AT+KUDPCLOSE=1,0\r";
int HL7800Udp::endPacket() {
  int ret = 0;
  char* rcv_buff;
  
  // AT+KUDPCLOSE=<session_id> [,<keep_cfg>]
  sprintf_P(cmd,PSTR("AT+KUDPCLOSE=%d,0\r"),1);
//  DBGSERIAL.print(F("@@@@@ HL7800Udp::endPacket() cmd="));
//  DBGSERIAL.println(cmd);
  rcv_buff = serial_hl78.WriteCommandAndReadResponse(cmd, ATRSP_OK, ATRSP_ERR, MODEM_CONNECT_TIMEOUT);
  if (rcv_buff == NULL) {
    DBGSERIAL.println(F("@@@@@ UDP setup fail!"));
  }
  ret = 1;
  return ret;
}

/**
 * UDP送信データ書き込み<br>
 * Writes UDP data to the remote connection. Must be wrapped between beginPacket() and endPacket(). <br>
 * beginPacket() initializes the packet of data, it is not sent until endPacket() is called. <br>
 * HL7800はバイト送信を未サポート<br>
 * @param data 送信データ
 * @return  1:成功, 0:失敗
 */
size_t HL7800Udp::write(uint8_t) {
  int ret = 0;
  DBGSERIAL.print(F("@@@@@ HL7800Udp::write(byte)"));
  return ret;
}
 
/**
 * UDP送信データ書き込み<br>
 * Writes UDP data to the remote connection. Must be wrapped between beginPacket() and endPacket(). <br>
 * beginPacket() initializes the packet of data, it is not sent until endPacket() is called. <br>
 * 指定されたデータ列をUDPで送信する<br>
 * @param buffer 送信データ列
 * @param size 送信データ数
 * @return  1:成功, 0:失敗
 */
//PROGMEM const char ATCMD_KUDPSND_HOST[]="AT+KUDPSND=1,\"%s\",%d,%d\r\n";
//PROGMEM const char ATCMD_KUDPSND_IP[]="AT+KUDPSND=1,\"%d.%d.%d.%d\",%d,%d\r\n";
size_t HL7800Udp::write(const uint8_t *buffer, size_t size) {
  int ret = 0;
  char* rcv_buff;
  int hostlen = 0;
  char write_cmd[size + 21]; // 書き込みデータ＋”\r\n\r\n--EOF--Pattern--" の配列を確保する

  if (_validHost) {
    hostlen = strlen(_targetHost);
  } else  if (_validIP) {
    hostlen = 15; //  IP:xxx.xxx.xxx.xxx
  } else {
    DBGSERIAL.println(F("not set HOST or IP!"));
    return ret;
  }
  char long_cmd[hostlen+25];

  // 送信コマンド+KUDPSNDを発行し接続待ち
  // AT+KUDPSND=<session_id>, <udp remote address>,<udp_port>, <ndata>
  // AT+KUDPSND=1,”213.41.22.60”,32,10
  if (_validHost) {
    sprintf_P(long_cmd,PSTR("AT+KUDPSND=1,\"%s\",%d,%d\r\n"),_targetHost, _targetPort, size);
  } else  if (_validIP) {
    sprintf_P(long_cmd,PSTR("AT+KUDPSND=1,\"%d.%d.%d.%d\",%d,%d\r\n"),_targetIP[0],_targetIP[1],_targetIP[2],_targetIP[3],_targetPort,size);
  } else {
    DBGSERIAL.println(F("not set HOST or IP!"));
    return ret;
  }
  DBGSERIAL.print(F("@@@@@ HL7800Udp::write() cmd="));
  DBGSERIAL.println(long_cmd);
  rcv_buff = serial_hl78.WriteCommandAndReadResponse(long_cmd, ATRSP_CONNECT, ATRSP_ERR, MODEM_CONNECT_TIMEOUT);
  if (rcv_buff == NULL) {
    DBGSERIAL.println(F("@@@@@ HL7800Udp::write() send fail!"));
    return ret;
  }

  // UDP送信データ部分を送信
  serial_hl78.WriteData(buffer, size);

  // 終端コード"--EOF--Pattern--"を発行し終了待ち
  rcv_buff = serial_hl78.WriteCommandAndReadResponse("--EOF--Pattern--", ATRSP_OK, ATRSP_ERR, MODEM_LONG_TIMEOUT);
  if (rcv_buff == NULL) {
    DBGSERIAL.println(F("@@@@@ HL7800Udp::write() send fail!"));
    return ret;
  }
  DBGSERIAL.println(F("@@@@@ HL7800Udp::write() send sucess."));
  ret = 1;
  return ret;
 }

/**
 * UDP受信待ち<br>
 * Checks for the presence of a UDP packet, and reports the size. parsePacket() must be called before reading the buffer <br>
 * モデムからのUDP受信通知を待ち<br>
 * @param なし
 * @return  取得したデータバイト数
 */
PROGMEM const char ATRSP_KUDP_DATA[]="+KUDP_DATA: 1,";
PROGMEM const char ATRSP_KUDP_NOTIF[]="+KUDP_NOTIF: 1,5";
PROGMEM const char ATRSP_KUDP_RCV[]="+KUDP_RCV: \"";
//PROGMEM const char ATCMD_KUDPRCV[]="AT+KUDPRCV=1,%d";
int HL7800Udp::parsePacket() {
  int ret = 0;
  char* rcv_buff;

  if (_remaining > 0) {
    // 受信済みのデータが残っている場合は、そのデータ数を返す
    return _remaining;  
  }
  
  // +KTCP_DATAを待つ
  // +KUDP_DATA: <session_id>,<ndata available>[,<udp remote address>,<udp remote port>,<data>]
  rcv_buff = serial_hl78.ReadResponse(ATRSP_KUDP_DATA, ATRSP_KUDP_NOTIF,  MODEM_CONNECT_TIMEOUT);  // 受信通知 / UDP connection error
  if (rcv_buff == NULL) {
    DBGSERIAL.print(F("UDP timeout"));
    // タイムアウトの場合受信バッファのゴミを捨てる
    rcv_buff = serial_hl78.ReadResponse("\r", ATRSP_ERR,  MODEM_SHORT_TIMEOUT);  
    return ret;
  }

  // データ長と通知コード終端を待つ
  rcv_buff = serial_hl78.ReadResponse("\n", ATRSP_KUDP_DATA,  MODEM_CONNECT_TIMEOUT);  // EOL / UDP connection error
  if (rcv_buff == NULL) {
    DBGSERIAL.print(F("UDP timeout"));
    // タイムアウトの場合受信バッファのゴミを捨てる
    rcv_buff = serial_hl78.ReadResponse("\r", ATRSP_ERR,  MODEM_SHORT_TIMEOUT);  
    return ret;
  }
  int len = atoi(rcv_buff); // 受信したデータ長を取得
  
  // UDPデータ受信コマンドを発行
  // AT+KUDPRCV= <session_id>, <ndata>
  sprintf_P(cmd,PSTR("AT+KUDPRCV=1,%d"),len);
  rcv_buff = serial_hl78.WriteCommandAndReadResponse(cmd, ATRSP_CONNECT, ATRSP_ERR, MODEM_CONNECT_TIMEOUT);
  if (rcv_buff == NULL) {
    DBGSERIAL.println(F("@@@@@ UDP recive fail!"));
    return ret;
  }

  // EOFまで受信
  rcv_buff = serial_hl78.ReadResponse(ATDATA_EOF, ATRSP_ERR,  MODEM_CONNECT_TIMEOUT);  
  if (rcv_buff == NULL) {
    DBGSERIAL.println(F("@@@@@ UDP read fail!(EOF not found)"));
    return ret;
  }

  // EOF文字列は不要なので削除
  char* end = strstr_P(rcv_buff,ATDATA_EOF);
  *end = NULL;
  int recived_size = strlen(rcv_buff);

  //　UDP受信バッファにコピーする
  memcpy(rxUdpBuff,rcv_buff,recived_size);
  memset(rxUdpBuff+recived_size,0x0,1);
  _remaining = recived_size;
  _offset = 0;

  // +KUDP_RCV:まで受信
  // +KUDP_RCV: <udp remote address>,<udp remote port>
  rcv_buff = serial_hl78.ReadResponse(ATRSP_KUDP_RCV, ATRSP_ERR,  MODEM_CONNECT_TIMEOUT);  
  if (rcv_buff == NULL) {
    DBGSERIAL.println(F("@@@@@ UDP read fail!(+KUDP_RCV: not found)"));
    return ret;
  }

  // ToDo: 文字列を解析し変数に代入
  _remoteIP = {0,0,0,0};
  _remotePort = 0;
  
  return recived_size;
}

/**
 * UDP有効データ確認<br>
 * Get the number of bytes (characters) available for reading from the buffer. This is data that's already arrived<br>
 * Checks for the presence of a UDP packet, and reports the size. parsePacket() must be called before reading the buffer <br>
 * 受信済みデータのデータサイズを返す<br>
 * @param なし
 * @return  受信済みデータバイト数
 */
int HL7800Udp::available() {
  return _remaining;
}

/**
 * UDPデータ取得<br>
 * Reads UDP data from the specified buffer. If no arguments are given, it will return the next character in the buffer. <br>
 * 受信済みデータのデータサイズを返す<br>
 * @param なし
 * @return  整数：有効データ, -1:失敗
 */
int HL7800Udp::read() {
  int ret = -1;
  if (_remaining > 0) {
    _remaining++;
    return rxUdpBuff[_offset++];
  }
  return ret;
}

/**
 * UDPデータ取得<br>
 * Reads UDP data from the specified buffer. If no arguments are given, it will return the next character in the buffer. <br>
 * 受信済みデータのデータサイズを返す<br>
 * @param buffer 受け取りデータバッファ
 * @param len 受信したいデータ数
 * @return  取得成功したデータバイト数
 */
int HL7800Udp::read(unsigned char* buffer, size_t len) {
  int ret = 0;
  if (_remaining > len) {
    ret = len;
  } else {
    ret = _remaining;
  }
  memcpy(buffer,rxUdpBuff,ret);
  _remaining = _remaining - ret;
  return len;
}

/**
 * UDPデータ廃棄
 * @return  1:成功, 0:失敗
 */
int HL7800Udp::peek() {
  int ret = 0;
  return ret;
}

/**
 * UDPデータ送信完了待ち<br>
 * write()で送信完了をチェックしているため、この関数は意味を持たない<br>
 * @return  1:成功, 0:失敗
 */
void HL7800Udp::flush() {
}
